/*
 * Copyright 2017 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.web.help;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Map;

import dk.vajhoej.vms.web.FunctionModule;
import dk.vajhoej.vms.web.Result;
import dk.vajhoej.vms.web.Util;

/**
 * Class HelpModule exposes VMS HELP via LBR$ routines.
 */
public class HelpModule implements FunctionModule {
	private static final String DEFAULT_TOPIC = "HELP";
	private static final String DELIM = "Additional information available:";
	private static String getText(String helpoutput) {
		String[] areas = helpoutput.split(DELIM);
		return areas[0];
	}
	private static String[] getSubTopics(String helpoutput) {
		String[] areas = helpoutput.split(DELIM);
		if(areas.length > 1) {
			return areas[1].replaceAll(" D=\\s+",  "").split("\\s+");
		} else {
			return new String[0];
		}
	}
	private static class HelpOutput {
		private String text;
		private String[] subTopics;
		public HelpOutput(String text, String[] subTopics) {
			this.text = text;
			this.subTopics = subTopics;
		}
		public String getText() {
			return text;
		}
		public String[] getSubTopics() {
			return subTopics;
		}
	}
	private static String toHTML(String txt) throws IOException {
		StringBuilder sb = new StringBuilder();
		BufferedReader br = new BufferedReader(new StringReader(txt));
		sb.append("<p>\r\n");
		String line;
		while((line = br.readLine()) != null) {
			if(line.trim().equals("")) {
				sb.append("</p>\r\n<p>\r\n");
			} else {
				sb.append(line);
				sb.append("\r\n");
			}
		}
		br.close();
		sb.append("</p>\r\n");
		return sb.toString();
	}
	private LBRWrapper lbr = new LBRWrapper();
	private HelpOutput getHelp(String[] topics) throws UnsupportedEncodingException {
		String helpoutput = lbr.get(topics);
		return new HelpOutput(getText(helpoutput), getSubTopics(helpoutput));
	}
	private Map<String,HelpOutput> cache = new HashMap<String,HelpOutput>();
	private HelpOutput getWithCache(String query) throws UnsupportedEncodingException {
		HelpOutput data;
		if(cache.containsKey(query)) {
			data = cache.get(query);
		} else {
			String[] topics = query.split("\\+");
			if(topics.length> 5) {
				return null;
			}
			data = getHelp(topics);
			cache.put(query, data);
		}
		return data;
	}
	private static boolean xcontains(String s, String subs) {
		s = s.toUpperCase();
		subs = subs.toUpperCase();
		if(s.endsWith(subs)) return true;
		if(s.contains(subs + "=")) return true;
		if(s.contains(subs + "[=")) return true;
		if(s.contains(subs + " (DEFAULT)")) return true;
		if(s.charAt(0) == '/' && s.contains(subs)) return true;
		if(s.charAt(0) == '/' && s.contains("/NO" + subs.substring(1))) return true;
		return false;
	}
	private String peek(String query, String subtopic, String link) throws IOException {
		String[] topics = query.split("\\+");
		HelpOutput data = getWithCache(link);
		String txt = data.getText();
		BufferedReader br = new BufferedReader(new StringReader(txt));
		String line;
	next:
		while((line = br.readLine()) != null) {
		    line = line.trim();
			if(line.length() == 0) continue next;
			if(xcontains(line, subtopic)) continue next;
			for(String topic : topics) {
				if(xcontains(line, topic)) continue next;
			}
			break;
		}
		br.close();
		if(line == null) return "";
		int ix = line.indexOf('.');
		if(ix < 0) {
			line = line + " ...";
			ix = line.length();
		}
		return line.substring(0, ix);
	}
	private String format(String query, HelpOutput data) throws IOException {
		String maintmpl = Util.loadText("main.tmpl");
		String linktmpl = Util.loadText("link.tmpl");
		String res = maintmpl;
		// main text
		res = res.replace("%TEXT%", data.getText()).replace("%HTML%", toHTML(data.getText())).replace("%TOPIC%", query.replace("+", " "));
		// sub topics
		StringBuilder sb = new StringBuilder();
		for(String subtopic : data.getSubTopics()) {
			if(subtopic.length() > 0) {
				String link;
				if(query.equals(DEFAULT_TOPIC)) {
					link = subtopic;
				} else {
					link = query + "+" + subtopic;
				}
				if(linktmpl.contains("%SUBTOPICPEEK%")) {
					sb.append(linktmpl.replace("%SUBTOPIC%", subtopic).replace("%LINK%", link).replace("%SUBTOPICPEEK%", peek(query, subtopic, link)));
				} else {
					sb.append(linktmpl.replace("%SUBTOPIC%", subtopic).replace("%LINK%", link));
				}
			}
		}
		res = res.replace("%SUBTOPICS%", sb.toString());
		// back link
		String backlink;
		int lastsep = query.lastIndexOf("+");
		if(lastsep > 0) {
			backlink = query.substring(0, lastsep);
		} else {
			backlink = DEFAULT_TOPIC;
		}
		res = res.replace("%BACKLINK%", linktmpl.replace("%SUBTOPIC%", "Back").replace("%LINK%", backlink).replace("%SUBTOPICPEEK%", ""));
		// top link
		res = res.replace("%TOPLINK%", linktmpl.replace("%SUBTOPIC%", "Top").replace("%LINK%", DEFAULT_TOPIC).replace("%SUBTOPICPEEK%", ""));
		//
		return  res;
	}
	public String getContext() {
		return "help";
	}
	public Result getResponse(String remurl, String query) {
		try {
			lbr.init();
			if(query == null || query.equals("")) query = DEFAULT_TOPIC;
			HelpOutput data = getWithCache(query);
			if(data == null) new Result(false, "Bad topics specified", null, null);
			String res = format(query, data);
			lbr.close();
			return new Result(true, null, "text/html", res.getBytes("UTF-8"));
		} catch (Exception e) {
			return new Result(false, "Error getting help information: " + e.getMessage(), null, null);
		}
	}
}
